/*
 *  This file is part of uIRCd. (https://git.redxen.eu/caskd/uIRCd)
 *  Copyright (c) 2019, 2020 Alex-David Denes
 *
 *  uIRCd is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  any later version.
 *
 *  uIRCd is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with uIRCd.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "buffer.h"

#include "logging.h"

#include <errno.h> // errno
#include <stdio.h> // fprintf()
#include <stdlib.h>
#include <string.h>    // strerror()
#include <sys/types.h> // size_t ssize_t
#include <unistd.h>    // write()

#define UIRC_IRCV3
#define UIRC_HELPERS
#include <uirc/uirc.h> // Tok_mesg()

ssize_t
get_buffer_line(char* buf)
{
	if (buf == NULL) return -1;
	char* ppoi;
	if ((ppoi = strchr(buf, '\n')) != NULL) {
		*ppoi = '\0';
		if (ppoi > buf && *(ppoi - 1) == '\r') *(ppoi - 1) = '\0';
		return ++ppoi - buf;
	}
	return 0;
}

ssize_t
flush_buffer(char* buf, size_t buflen, int fd)
{
	ssize_t res;
	char*   pos = buf;
	for (;;) {
		if ((size_t)(res = write(fd, pos, buflen - (size_t)(pos - buf))) != buflen - (size_t)(pos - buf)) {
			if (res == -1 && errno != EINTR) {
				LOG(LOG_WARN, "Couldn't flush buffer to fd %i. " ERRNOFMT, fd, strerror(errno), errno);
				return -1;
			}
			pos += res;
			LOG(LOG_DEBUG, "Wrote %lu bytes to fd %i.", res, fd);
		} else
			return res;
	}
}

ssize_t
read_buffer(Buffer_Info* buf)
{
	ssize_t brd;
	if ((brd = read(buf->fd, buf->buffer + buf->append_pos, buf->csize - buf->append_pos - 1)) > 0) {
		*(buf->buffer + (buf->append_pos += (size_t) brd)) = '\0';
		LOG(LOG_DEBUG, "Read %li bytes from socket buffer. Now appending at %li with \"%s\" so far.", brd, buf->append_pos, buf->buffer);
	} else if (brd == -1) {
		if (errno == EAGAIN || errno == EWOULDBLOCK || errno == EINTR) return 0;
		LOG(LOG_WARN, "Failed to read inbound traffic on fd %i. " ERRNOFMT, buf->fd, strerror(errno), errno);
	}
	return brd;
}

int
reset_buffer(Buffer_Info* buf, short clr_fd)
{
	if (clr_fd) {
		if (clr_fd == 2) close(buf->fd);
		buf->fd = -1;
	}
#ifdef UIRCD_RELAXED_RFC
	free(buf->buffer);
	buf->buffer = NULL;
	buf->csize  = 0;
#else  /* UIRCD_RELAXED_RFC */
	*buf->buffer = '\0';
	buf->csize   = UIRCD_LIMITS_LINE + 1;
#endif /* UIRCD_RELAXED_RFC */
	buf->append_pos = 0;
	return 1;
}

// TODO: Buffer allocators and resizers
